/*
 *  
 *  Copyright (c) 2004 Steve Slaven, All Rights Reserved.
 *  
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *  MA 02111-1307 USA
 *  
*/
package net.hoopajoo.cheeselab3;

import java.util.*;

class PathFinderBFS extends PathFinder {
    private long stat_loop_iter = 0,
                 stat_found_better_path = 0;
    private boolean avoid_water = false;
    private int max_iterations = 0;

    class PathElement {
        CheeseRoomData room;
        String path;
    }

    // All the pathelements are in the hash, the vector just has integer ids
    // that must be resolved in the has (so we can easily discard redundant data)
    protected String perform_find_path( CheeseRoomData from, CheeseRoomData to ) {
        PathElement ret = new PathElement();
        ret.room = from;
        ret.path = "";

        Vector to_explore = new Vector();
        to_explore.add( new Integer( from.roomid ) );

        HashMap elements = new HashMap();
        elements.put( new Integer( from.roomid ), ret );

        boolean looping = true;
        while( looping ) {
            stat_loop_iter++;

            Integer roomid = (Integer)to_explore.get( 0 );
            to_explore.remove( 0 );

            ret = (PathElement)elements.get( roomid );
            if( ret == null ) {
                // Logic error?
                set_error( "NULL element in to_explore?" );
                return( null );
            }

            if( ret.room.roomid == to.roomid ) {
                // Found it
                looping = false;
            } else {
                // Expand each exit in to a new pathelement, if
                // there was nothing in the elements before *or* the
                // path is shorter, then add this room to the to_explore
                // list and update elements
                update_elements( elements, to_explore, ret, "n", ret.room.exit_n );
                update_elements( elements, to_explore, ret, "s", ret.room.exit_s );
                update_elements( elements, to_explore, ret, "e", ret.room.exit_e );
                update_elements( elements, to_explore, ret, "w", ret.room.exit_w );
                update_elements( elements, to_explore, ret, "u", ret.room.exit_u );
                update_elements( elements, to_explore, ret, "d", ret.room.exit_d );
            }

            // Exit conditions
            if( to_explore.isEmpty() ) {
                looping = false;
                set_error( "Unable to find a path, all routes exhausted" );
            }

            if( looping && max_iterations != 0 && stat_loop_iter > max_iterations ) {
                set_error( "Hit max_iterations" );
                ret.path = null;
                looping = false;
            }

        }

        return( ret.path );
    }

    private void update_elements( HashMap elements, Vector to_explore,
                                  PathElement last, String direction, int toid ) {
        // Make sure it's an ok direction
        if( toid < 1 )
            return;

        Integer roomid = new Integer( toid );

        // Check if it's in the elements already
        PathElement next = (PathElement)elements.get( roomid );
        if( next == null ) {
            // Create it
            next = new PathElement();
            next.path = last.path + direction;
            next.room = get_room( toid );

            // If not water...
            if( avoid_water )
                if( next.room.terrain.equals( "U" )
                        || next.room.terrain.equals( "w" )
                        || next.room.terrain.equals( "W" ) )
                    next = null;

            if( next != null ) {
                // Store it in to_explore and elements
                to_explore.add( roomid );
                elements.put( roomid, next );
            }
        } else {
            // If the last path elements cost is lower, then supercede this one
            if( next.path.length() > last.path.length() + 1 ) {
                stat_found_better_path++;
                to_explore.add( roomid );
                next.path = last.path + direction;
                elements.put( roomid, next );
            }
        }
    }

    public void print_stats() {
        System.out.println( "Loop iterations: " + stat_loop_iter );
        System.out.println( "Max iterations: " + max_iterations );
        System.out.println( "Better paths found: " + stat_found_better_path );
        System.out.println( "Avoid water: " + avoid_water );
        super.print_stats();
    }

    public void set_option( String key, String value ) {
        if( key.equals( "avoid_water" ) ) {
            avoid_water = value_to_boolean( value );
        } else if( key.equals( "max_iterations" ) ) {
            max_iterations = value_to_int( value );
        } else {
            super.set_option( key, value );
        }
    }
}
